from typing import Dict, List

from routing.app_settings import corptools_active
from routing.models import SolarSystemConnection, TrigInvasion

if corptools_active():
    from corptools.models import MapJumpBridge


def include_corptools_jumpbridges():
    edges = []
    if corptools_active():
        for jb in MapJumpBridge.objects.values_list("from_solar_system_id", "to_solar_system_id").all():
            edges.append((
                jb.fromsolarsystem, jb.tosolarsystem,
                {'prefer_shortest': 1.0, 'prefer_safest': 1.0, 'prefer_less_safe': 1.0}))
    else:
        return edges

    return edges


def include_eve_scout(system: str = "thera") -> List[Dict]:
    edges = []
    if system == "thera":
        pass
    elif system == "turnur":
        pass
    return edges
    # return [(30100000, 30003841, {'prefer_shortest': 1.0, 'prefer_safest': 1.0, 'prefer_less_safe': 1.0})]


def avoid_edencom():
    edges = []
    for system in TrigInvasion.objects.filter(status__in=["fortress", "edencom_minor_victory"]).values_list("tosolarsystem", flat=True).all():
        for connection in SolarSystemConnection.objects.filter(to_solar_system=system):
            edges.append((
                connection.fromsolarsystem, connection.tosolarsystem,
                {'prefer_shortest': 50000.0, 'prefer_safest': 50000.0, 'prefer_less_safe': 50000.0}))

    return edges



from django.apps import AppConfig

from . import __version__


class RoutingConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = "routing"
    verbose_name = f'AA Routing v{__version__}'



from networkx import to_dict_of_dicts

from django.core.management.base import BaseCommand

from routing.graph import build


class Command(BaseCommand):
    help = ""

    def handle(self, *args, **options):

        G = build()

        with open("routing_output.txt", "w") as f:
            f.write(str(to_dict_of_dicts(G)))



# Generated by Django 4.2.6 on 2023-11-02 08:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('routing', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='solarsystemconnection',
            name='prefer_less_safe',
            field=models.FloatField(default=1, verbose_name='Weighting for Prefer Less Safe Routing'),
        ),
        migrations.AddField(
            model_name='solarsystemconnection',
            name='prefer_safest',
            field=models.FloatField(default=1, verbose_name='Weighting for Prefer Safest Routing'),
        ),
        migrations.AddField(
            model_name='solarsystemconnection',
            name='prefer_shortest',
            field=models.FloatField(default=1, verbose_name='Weighting for Prefer Shortest Routing'),
        ),
    ]



from typing import List

from networkx import (
    astar_path, bellman_ford_path, dijkstra_path, single_source_dijkstra_path,
)

from .graph import build


def route_path(source: int, destination: int, mode="prefer_shortest", algorithm="astar", edges: List = [], static_cache: bool = False) -> List[int]:
    """_summary_

    Args:
        source (int): From Solar System ID
        destination (int): To Solar System ID
        mode (str, optional): Weighting mode Defaults to "prefer_shortest". ["prefer_shortest", "prefer_safest", "prefer_less_safe"]
        algorithm (str, optional): Routing Function to use. Defaults to "astar" ["astar", "dijkstra", "bellman_ford"]
        edges (list, optional): Extra edges to load, example [(30100000, 30003841, {'prefer_shortest': 1.0, 'prefer_safest': 1.0, 'prefer_less_safe': 1.0}),]. Defaults to [].
        static_cache (bool, optional): Use Pregenerated Cache. Defaults to False.

    Returns:
        [int]: list of integer Solar System IDs
    """
    G = build(static_cache)
    if edges is not {}:
        G.add_edges_from(edges)

    if algorithm == "astar":
        return astar_path(build(static_cache), source, destination, weight=mode)
    elif algorithm == "dijkstra":
        return dijkstra_path(build(static_cache), source, destination, weight=mode)
    elif algorithm == "bellman_ford":
        return bellman_ford_path(build(static_cache), source, destination, weight=mode)
    else:
        return astar_path(build(static_cache), source, destination, weight=mode)


def route_length(source: int, destination: int, mode="prefer_shortest", algorithm="astar", edges: List = [], static_cache: bool = False) -> int:
    """_summary_

    Args:
        source (int): From Solar System ID
        destination (int): To Solar System ID
        mode (str, optional): Weighting mode Defaults to "prefer_shortest". ["prefer_shortest", "prefer_safest", "prefer_less_safe"]
        algorithm (str, optional): Routing Function to use. Defaults to "astar" ["dijkstra", "bellman_ford"]
        edges (list, optional): Extra edges to load, example [(30100000, 30003841, {'prefer_shortest': 1.0, 'prefer_safest': 1.0, 'prefer_less_safe': 1.0})]. Defaults to [].
        static_cache (bool, optional): Use Pregenerated Cache. Defaults to False.

    Returns:
        int : the number of JUMPS, this will be one shorter than route_path, as it includes the source system.
    """

    return len(route_path(source, destination, mode, algorithm, edges, static_cache)) - 1


def systems_range(source: int, range: int, mode="prefer_shortest", edges: list = [], static_cache: bool = False) -> List:
    list = [key for key in single_source_dijkstra_path(build(static_cache), source, range, weight=mode)]
    list.remove(source)
    return list



import bz2
import csv
import os

import requests

from routing.models import SolarSystem, SolarSystemConnection, TrigInvasion


def pull_data_solarsystems():
    mapjumps_url = 'https://www.fuzzwork.co.uk/dump/latest/mapSolarSystems.csv.bz2'
    SolarSystem.objects.all().delete()

    connections = []

    with open('mapSolarSystems.csv.bz2', 'wb') as f:
        f.write(requests.get(mapjumps_url).content)

    open('mapSolarSystems.csv', 'wb').write(
        bz2.open('mapSolarSystems.csv.bz2', 'rb').read())

    with open('mapSolarSystems.csv', encoding='UTF-8') as f:
        csv_file = csv.reader(f)
        next(csv_file)  # skip headers
        for row in csv_file:
            connections.append(SolarSystem(id=row[2], security_status=row[21]))

    SolarSystem.objects.bulk_create(connections, batch_size=500, ignore_conflicts=True)


def pull_data_connections():
    mapjumps_url = 'https://www.fuzzwork.co.uk/dump/latest/mapSolarSystemJumps.csv.bz2'
    SolarSystemConnection.objects.all().delete()

    connections = []

    with open('mapSolarSystemJumps.csv.bz2', 'wb') as f:
        f.write(requests.get(mapjumps_url).content)

    open('mapSolarSystemJumps.csv', 'wb').write(
        bz2.open('mapSolarSystemJumps.csv.bz2', 'rb').read())

    with open('mapSolarSystemJumps.csv', encoding='UTF-8') as f:
        csv_file = csv.reader(f)
        next(csv_file)  # skip headers
        for row in csv_file:
            # fromsolarsystem = SolarSystem.objects.get(id=row[2])
            tosolarsystem = SolarSystem.objects.get(id=row[3])

            connections.append(SolarSystemConnection(
                fromsolarsystem_id=row[2],
                tosolarsystem_id=row[3],
                prefer_shortest=1,
                prefer_safest=1 if tosolarsystem.security_status >= 0.45 else 50000.0,  # High Sec
                prefer_less_safe=1 if tosolarsystem.security_status < 0.45 else 50000.0,  # Low/Null
            ))

    SolarSystemConnection.objects.bulk_create(connections, batch_size=500, ignore_conflicts=True)


def import_trig_data():
    TRIG_CSV_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'trigInvasionSystems.csv')
    trig_systems = []
    with open(TRIG_CSV_PATH, encoding='UTF-8') as f:
        csv_file = csv.reader(f)
        next(csv_file)  # skip headers
        for row in csv_file:
            trig_systems.append(TrigInvasion(
                system_id=row[0],
                status=row[1]
            ))

    TrigInvasion.objects.bulk_create(trig_systems, batch_size=500, ignore_conflicts=True)



from networkx import DiGraph

from routing.models import SolarSystem, SolarSystemConnection

from .static_data import precomputed_graph


def build(static_cache: bool = False) -> DiGraph:
    if static_cache:
        G = DiGraph(precomputed_graph)
        return G
    else:
        G = DiGraph()
        for node in SolarSystem.objects.values_list("id", "security_status"):
            G.add_node(node[0], security_status=node[1], type="stargate")

        for edge in SolarSystemConnection.objects.values_list(
            "fromsolarsystem",
            "tosolarsystem",
            "prefer_shortest",
            "prefer_safest",
            "prefer_less_safe"
        ).all():
            G.add_edge(edge[0], edge[1], prefer_shortest=edge[2], prefer_safest=edge[3], prefer_less_safe=edge[4])

        return G



from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _


class SolarSystem(models.Model):
    id = models.PositiveIntegerField(_("Solar System ID"), primary_key=True)
    security_status = models.FloatField(
        _("Security Status"),
        validators=[MinValueValidator(0.0), MaxValueValidator(1.0)])

    class Meta:
        verbose_name = _("Solar System")
        verbose_name_plural = _("Solar Systems")


class SolarSystemConnection(models.Model):
    fromsolarsystem = models.ForeignKey(
        SolarSystem,
        verbose_name=_("From Solar System"),
        on_delete=models.CASCADE,
        related_name="+")
    tosolarsystem = models.ForeignKey(
        SolarSystem,
        verbose_name=_("To Solar System"),
        on_delete=models.CASCADE,
        related_name="+")

    prefer_shortest = models.FloatField(_("Weighting for Prefer Shortest Routing"), default=1)
    prefer_safest = models.FloatField(_("Weighting for Prefer Safest Routing"), default=1)
    prefer_less_safe = models.FloatField(_("Weighting for Prefer Less Safe Routing"), default=1)

    class Meta:
        verbose_name = _("Solar System Connections")
        verbose_name_plural = _("Solar System Connection")
        unique_together = ["fromsolarsystem", "tosolarsystem"]


class TrigInvasion(models.Model):

    class Status(models.TextChoices):
        """"
        The system status after the Triglavian Invasion
        """
        EDENCOM_MINOR_VICTORY = 'edencom_minor_victory', _("EDENCOM Minor Victory")
        FINAL_LIMINALITY = 'final_liminality', _("Final Liminality")
        FORTRESS = 'fortress', _("Fortress")
        TRIGLAVIAN_MINOR_VICTORY = 'triglavian_minor_victory', _("Triglavian Minor Victory")

    system = models.OneToOneField(
        SolarSystem,
        verbose_name=_("Solar System"),
        on_delete=models.CASCADE,
        related_name="+",
        primary_key=True)

    status = models.CharField(_("Mass Remaining"), max_length=50, choices=Status.choices)

    class Meta:
        verbose_name = _("Triglavian/Edencom Invasion")
        verbose_name_plural = _("Triglavian/Edencom Invasions")



# Generated by Django 4.2.6 on 2023-11-02 13:01

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('routing', '0002_solarsystemconnection_prefer_less_safe_and_more'),
    ]

    operations = [
        migrations.CreateModel(
            name='TrigInvasion',
            fields=[
                ('system', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, primary_key=True, related_name='+', serialize=False, to='routing.solarsystem', verbose_name='Solar System')),
                ('status', models.CharField(choices=[('edencom_minor_victory', 'EDENCOM Minor Victory'), ('final_liminality', 'Final Liminality'), ('fortress', 'Fortress'), ('triglavian_minor_victory', 'Triglavian Minor Victory')], max_length=50, verbose_name='Mass Remaining')),
            ],
            options={
                'verbose_name': 'Triglavian/Edencom Invasion',
                'verbose_name_plural': 'Triglavian/Edencom Invasions',
            },
        ),
    ]



# Generated by Django 4.2.6 on 2023-11-02 07:44

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SolarSystem',
            fields=[
                ('id', models.PositiveIntegerField(primary_key=True, serialize=False, verbose_name='Solar System ID')),
                ('security_status', models.FloatField(validators=[django.core.validators.MinValueValidator(0.0), django.core.validators.MaxValueValidator(1.0)], verbose_name='Security Status')),
            ],
            options={
                'verbose_name': 'Solar System',
                'verbose_name_plural': 'Solar Systems',
            },
        ),
        migrations.CreateModel(
            name='SolarSystemConnection',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('fromsolarsystem', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='routing.solarsystem', verbose_name='From Solar System')),
                ('tosolarsystem', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='+', to='routing.solarsystem', verbose_name='To Solar System')),
            ],
            options={
                'verbose_name': 'Solar System Connections',
                'verbose_name_plural': 'Solar System Connection',
                'unique_together': {('fromsolarsystem', 'tosolarsystem')},
            },
        ),
    ]



from django.test import TestCase

from routing.models import SolarSystem, SolarSystemConnection


class SolarSystemModelTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Set up non-modified objects used by all test methods
        SolarSystem.objects.create(id=1, security_status=0.5)

    def test_security_status_label(self):
        solarsystem = SolarSystem.objects.get(id=1)
        field_label = solarsystem._meta.get_field('security_status').verbose_name
        self.assertEqual(field_label, 'Security Status')

    def test_security_status_max_value(self):
        solarsystem = SolarSystem.objects.get(id=1)
        max_value = solarsystem._meta.get_field('security_status').validators[1].limit_value
        self.assertEqual(max_value, 1.0)


class SolarSystemConnectionModelTest(TestCase):
    @classmethod
    def setUpTestData(cls):
        # Set up non-modified objects used by all test methods
        SolarSystem.objects.create(id=1, security_status=0.5)
        SolarSystem.objects.create(id=2, security_status=0.7)
        SolarSystemConnection.objects.create(
            fromsolarsystem_id=1,
            tosolarsystem_id=2,
            prefer_shortest=1,
            prefer_safest=1,
            prefer_less_safe=1)

    def test_prefer_shortest_default(self):
        connection = SolarSystemConnection.objects.get(id=1)
        default_value = connection._meta.get_field('prefer_shortest').default
        self.assertEqual(default_value, 1)

    def test_prefer_safest_default(self):
        connection = SolarSystemConnection.objects.get(id=1)
        default_value = connection._meta.get_field('prefer_safest').default
        self.assertEqual(default_value, 1)

    def test_prefer_less_safe_default(self):
        connection = SolarSystemConnection.objects.get(id=1)
        default_value = connection._meta.get_field('prefer_less_safe').default
        self.assertEqual(default_value, 1)
